' ======================================================================
'
' File:      MINIMAX.BAS
' Purpose:   A Didactic Chess Program
' Project:   Minimax in BASIC
' Compiler:  Visual Basic for DOS
' Authors:   D. Steinwender, Ch. Donninger
' Date:      May 1, 1995
'
' ----------------------------------------------------------------------
'
' FreeBASIC adaptation: S. Budinov
' CECP protocol implementation: R. Chastain
' http://www.freebasic.net/forum/viewtopic.php?f=8&t=19296
'
' ======================================================================

#include "crt/stdio.bi"
#include "file.bi"
#include "version.bi"

declare sub Initialize()
declare sub CommandLoop()
declare function AlphaBeta(AAlpha as short, ABeta as short, Distance as short) as short
declare sub GenerateMoves(AllMoves as short)
declare sub PerformMove(CurrMove as short)
declare sub TakeBackMove(CurrMove as short)
declare sub CopyMainVariant(CurrMove as short)
declare function AttackingField(AField as short, Side as short) as short
declare function AssessPosition(AAlpha as short, ABeta as short, Side as short) as short
declare sub InitAssessment()
declare function DisplayBoard() as string
declare sub ComputerMove()
declare sub InitGameTree()
declare sub DisplayMove(CurrMove as short)
declare function FieldNotation(AField as short) as string
declare function FieldNumber(AField as string) as short
declare sub SavePromotion(AFrom as short, ATo as short)
declare sub SaveMove(AFrom as short, ATo as short)
declare sub SaveCaptureMove(AFrom as short, ATo as short)
declare sub SaveEnPassantMove(AFrom as short, ATo as short, AEP as short)
declare function NextBestMove() as short
declare function BPAssessment(AField as short, AColumn as short, ARow as short, ADeveloped as short) as short
declare function WPAssessment(AField as short, AColumn as short, ARow as short, ADeveloped as short) as short
declare function InputMove(Move as string) as short

const CBoardDim as short = 119
const CMaxDepth as short = 19
const CMoveDirections as short = 15
const PieceTypes as short = 6
const CMoveStackDim as short = 1000
const BK as short = -6
const BQ as short = -5
const BN as short = -4
const BB as short = -3
const BR as short = -2
const BP as short = -1
const Nil as short = 0
const WP as short = 1
const WR as short = 2
const WB as short = 3
const WN as short = 4
const WQ as short = 5
const WK as short = 6
const Edge as short = 100
const MatP as short = 100
const MatR as short = 500
const MatB as short = 350
const MatN as short = 325
const MatQ as short = 900
const MatK as short = 0
const MateValue as short = 32000
const MaxPos as short = MatB
const MainVariantBonus as short = 500
const Killer1Bonus as short = 250
const Killer2Bonus as short = 150
const MaterialSum as short = 4 * (MatR + MatB + MatN) + (2 * MatQ)
const EndgameMaterial as short = 4 * MatR + 2 * MatB
const A1 as short = 21
const B1 as short = 22
const C1 as short = 23
const D1 as short = 24
const E1 as short = 25
const F1 as short = 26
const G1 as short = 27
const H1 as short = 28
const C2 as short = 33
const H2 as short = 38
const A3 as short = 41
const C3 as short = 43
const D3 as short = 44
const E3 as short = 45
const A6 as short = 71
const C6 as short = 73
const D6 as short = 74
const E6 as short = 75
const H6 as short = 78
const A7 as short = 81
const C7 as short = 83
const H7 as short = 88
const A8 as short = 91
const B8 as short = 92
const C8 as short = 93
const D8 as short = 94
const E8 as short = 95
const F8 as short = 96
const G8 as short = 97
const H8 as short = 98
const RowA as short = 1
const BRow as short = 2
const CRow as short = 3
const DRow as short = 4
const ERow as short = 5
const FRow as short = 6
const GRow as short = 7
const HRow as short = 8
const Column1 as short = 2
const Column2 as short = 3
const Column3 as short = 4
const Column4 as short = 5
const Column5 as short = 6
const Column6 as short = 7
const Column7 as short = 8
const Column8 as short = 9
const NoCastlingMove as short = 0
const ShortCastlingMove as short = 1
const LongCastlingMove as short = 2
const White as short = 1
const Black as short = -1
#undef true
#undef false
const true = 1
const false = 0
const Legal as short = 1
const Illegal as short = 0

type MoveType
  FromField as short
  ToField as short
  CapturedPiece as short
  PromotedPiece as short
  CastlingNr as short
  EpField as short
  Value as short
end type

type PieceOffsetType
  Start as short
  Ends as short
  Longpaths as short
end type

type BothColourTypes
  White as short
  Black as short
end type

type FromToType
  FromField as short
  ToField as short
end type

type KillerType
  Killer1 as FromToType
  Killer2 as FromToType
end type

dim shared Board(CBoardDim) as short
dim shared EpField(CMaxDepth) as short
dim shared MoveStack(CMoveStackDim) as MoveType
dim shared MoveControl(H8) as short
dim shared Castling(2) as short
dim shared Index as short
dim shared StackLimit(CMaxDepth) as short
dim shared MVar(CMaxDepth, CMaxDepth) as FromToType
dim shared KillerTab(CMaxDepth) as KillerType
dim shared PawnControlled(CBoardDim) as BothColourTypes
dim shared Pawns(HRow + 1) as BothColourTypes
dim shared Rooks(HRow + 1) as BothColourTypes
dim shared Mobility(CMaxDepth) as short
dim shared ToField(CMaxDepth) as short
dim shared WKing as short
dim shared BKing as short
dim shared MaterialBalance(CMaxDepth) as short
dim shared MaterialTotal(CMaxDepth) as short
dim shared SideToMove as short
dim shared ComputerSide as short
dim shared PlayerPlayer as short
dim shared MinDepth as short
dim shared MaxExtension as short
dim shared Depth as short
dim shared NodeCount as long
dim shared LastMove as short
dim shared InCheck as short
dim shared MoveCount as short
dim shared IsWhiteLast as short
dim shared InitialPosition(CBoardDim) as short

for i as short = 0 to CBoardDim
  read InitialPosition(i)
next

data 100,100,100,100,100,100,100,100,100,100
data 100,100,100,100,100,100,100,100,100,100
data 100, 2 , 4 , 3 , 5 , 6 , 3 , 4 , 2 ,100
data 100, 1 , 1 , 1 , 1 , 1 , 1 , 1 , 1 ,100
data 100, 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 ,100
data 100, 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 ,100
data 100, 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 ,100
data 100, 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 ,100
data 100,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,100
data 100,-2 ,-4 ,-3 ,-5 ,-6 ,-3 ,-4 ,-2 ,100
data 100,100,100,100,100,100,100,100,100,100
data 100,100,100,100,100,100,100,100,100,100

dim shared GOffset(CMoveDirections) as short

GOffset(0)  =  -9
GOffset(1)  = -11
GOffset(2)  =   9
GOffset(3)  =  11
GOffset(4)  =  -1
GOffset(5)  =  10
GOffset(6)  =   1
GOffset(7)  = -10
GOffset(8)  =  19
GOffset(9)  =  21
GOffset(10) =  12
GOffset(11) =  -8
GOffset(12) = -19
GOffset(13) = -21
GOffset(14) = -12
GOffset(15) =   8

dim shared FigOffset(PieceTypes) as PieceOffsetType

FigOffset(Nil).Start = 0
FigOffset(Nil).Ends = 0
FigOffset(Nil).Longpaths = false
FigOffset(WP).Start = -1
FigOffset(WP).Ends = -1
FigOffset(WP).Longpaths = false
FigOffset(WR).Start = 4
FigOffset(WR).Ends = 7
FigOffset(WR).Longpaths = true
FigOffset(WB).Start = 0
FigOffset(WB).Ends = 3
FigOffset(WB).Longpaths = true
FigOffset(WN).Start = 8
FigOffset(WN).Ends = 15
FigOffset(WN).Longpaths = false
FigOffset(WQ).Start = 0
FigOffset(WQ).Ends = 7
FigOffset(WQ).Longpaths = true
FigOffset(WK).Start = 0
FigOffset(WK).Ends = 7
FigOffset(WK).Longpaths = false

dim shared CenterTable(H8) as short

for i as short = 0 to H8
  read CenterTable(i)
next

data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 4, 0, 8,12,12, 8, 0, 4, 0
data 0, 4, 8,12,16,16,12, 8, 4, 0
data 0, 8,12,16,20,20,16,12, 8, 0
data 0,12,16,20,24,24,20,16,12, 0
data 0,12,16,20,24,24,20,16,12, 0
data 0, 8,12,16,20,20,16,12, 8, 0
data 0, 4, 8,12,16,16,12, 8, 4, 0
data 0, 4, 0, 8,12,12, 8, 0, 4

dim shared WPFieldValue(H7) as short

for i as short = 0 to H7
  read WPFieldValue(i)
next

data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 4, 4, 0, 0, 0, 6, 6, 6, 0
data 0, 6, 6, 8, 8, 8, 4, 6, 6, 0
data 0, 8, 8,16,22,22, 4, 4, 4, 0
data 0,10,10,20,26,26,10,10,10, 0
data 0,12,12,22,28,28,14,14,14, 0
data 0,18,18,28,32,32,20,20,20

dim shared BPFieldValue(H7) as short

for i as short = 0 to H7
  read BPFieldValue(i)
next

data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
data 0,18,18,28,32,32,20,20,20, 0
data 0,12,12,22,28,28,14,14,14, 0
data 0,10,10,20,26,26,10,10,10, 0
data 0, 8, 8,16,22,22, 4, 4, 4, 0
data 0, 6, 6, 8, 8, 8, 4, 6, 6, 0
data 0, 4, 4, 0, 0, 0, 6, 6, 6, 0

dim shared PieceMaterial(PieceTypes) as short

PieceMaterial(Nil) = 0
PieceMaterial(WP) = MatP
PieceMaterial(WR) = MatR
PieceMaterial(WB) = MatB
PieceMaterial(WN) = MatN
PieceMaterial(WQ) = MatQ
PieceMaterial(WK) = MatK

dim shared FigSymbol(PieceTypes) as string * 1

FigSymbol(Nil) = "."
FigSymbol(WP) = "P"
FigSymbol(WR) = "R"
FigSymbol(WB) = "B"
FigSymbol(WN) = "N"
FigSymbol(WQ) = "Q"
FigSymbol(WK) = "K"

dim shared appLog as integer
appLog = freefile
open "minimax.log" for output as #appLog
#define LogLn(x) print #appLog, x : fflush(cast(FILE ptr, fileattr(appLog, fbFileAttrHandle)))
LogLn("** " & PROGRAM_INFO)

sub Send(byref AMessage as string)
  print(AMessage)
  'fflush(STDOUT)
  LogLn("<- " & AMessage)
end sub

' ----------------------------------------------------------------------

Initialize
CommandLoop

' ----------------------------------------------------------------------

close #appLog

sub Initialize
  dim i as short
  MoveCount = 0
  for i = 0 to CBoardDim
    Board(i) = InitialPosition(i)
  next i
  WKing = E1
  BKing = E8
  for i = 0 to 2
    Castling(i) = false
  next i
  for i = A1 to H8
    MoveControl(i) = 0
  next i
  EpField(0) = Illegal
  MaterialTotal(0) = MaterialSum
  MaterialBalance(0) = 0
  PlayerPlayer = false
  StackLimit(0) = 0
  MinDepth = 4
  Depth = 0
  SideToMove = White
  ComputerSide = Nil
end sub

sub CommandLoop
  dim as short Ends
  dim as string FullCommand, ShortCommand
  dim i as integer
  do
    input "", FullCommand
    LogLn("-> " & FullCommand)
    
    i = instr(FullCommand, " ")
    if i > 0 then
      ShortCommand = mid(FullCommand, 1, i - 1)
    else
      ShortCommand = FullCommand
    end if
    
    select case ShortCommand
      case "accepted"
      case "display-assessment"
        InitGameTree
        print("# Assessment = " & AssessPosition(-1 * MateValue, MateValue, SideToMove))
      case "display-board"
        print(DisplayBoard())
      case "easy"
      case "force"
        ComputerSide = Nil
      case "go"
        ComputerSide = SideToMove
      case "hard"
      case "level"
      case "new"
        Initialize
      case "otim"
      case "post"
      case "protover"
        Send("feature ping=0 setboard=0 usermove=0 memory=0 sigint=0 sigterm=0 myname=""" & PROGRAM_NAME & " " & PROGRAM_VERSION & """")
        Send("feature done=1")
      case "quit"
        Ends = true
      case "random"
      case "rejected"
      case "result"
      case "time"
      case "xboard"
      case else
        if InputMove(FullCommand) = false then
          LogLn("** Illegal move or unknown command: " & FullCommand)
        end if
    end select
    
    if ComputerSide = SideToMove then
      ComputerMove
    end if
    
    sleep(10, 1)
  loop while Ends = false
end sub

function AlphaBeta(AAlpha as short, ABeta as short, Distance as short) as short
  NodeCount = NodeCount + 1
  MVar(Depth, Depth).FromField = 0
  dim as short BestValue, i, Value, Check
  Value = AssessPosition(AAlpha, ABeta, SideToMove)
  Check = InCheck
  dim as short Condition1 = (Check = true) and (Depth + Distance < MaxExtension + 1)
  dim as short Condition2 = (Depth >= 2) and (Depth + Distance < MaxExtension)
  Condition2 = Condition2 and (ToField(Depth) = ToField(Depth - 1))
  Condition2 = Condition2 and (Value >= AAlpha - 150) and (Value <= ABeta + 150)
  if Condition1 or Condition2 then Distance = Distance + 1
  if (Distance < -5) or (Value = MateValue - Depth) or (Depth >= CMaxDepth) then
    AlphaBeta = Value
    exit function
  end if
  if (Value >= ABeta) and (Distance + Check <= 1) then
    AlphaBeta = Value
    exit function
  end if
  GenerateMoves(Distance)
  if Distance > 0 then
    BestValue = -MateValue
  else
    BestValue = Value
  end if
  i = NextBestMove
  do while i >= 0
    PerformMove(i)
    Value = -1 * AlphaBeta(-1 * ABeta, -1 * AAlpha, Distance - 1)
    TakeBackMove(i)
    if Value > BestValue then
      BestValue = Value
      if Value >= ABeta then
      if Distance > 0 then CopyMainVariant(i)
      goto Done
    end if
    if Value > AAlpha then
      if Distance > 0 then CopyMainVariant(i)
      AAlpha = Value
    end if
    end if
    i = NextBestMove
  loop
  Done:
  if (Value >= ABeta) and (i >= 0) then
    KillerTab(Depth).Killer2 = KillerTab(Depth).Killer1
    KillerTab(Depth).Killer1.FromField = MoveStack(i).FromField
    KillerTab(Depth).Killer1.ToField = MoveStack(i).ToField
  end if
  if BestValue = -(MateValue - (Depth + 1)) then
    if Check = false then
      AlphaBeta = 0
      exit function
    end if
  end if
  AlphaBeta = BestValue
end function

function AssessPosition(AAlpha as short, ABeta as short, Side as short ) as short
  dim as short Value, PosValue, i, j, k, Feld
  dim as short wBishlop, bBishlop, PawnCount
  dim as short MatNum, wRookon7, bRookon2, wDeveloped, bDeveloped
  dim as short ColumnnDiff, RownDiff, Bed1, Bed2, Bed3
  if Side = White then
    if AttackingField(BKing, White) = true then
      AssessPosition = MateValue - Depth
      exit function
    end if
    InCheck = AttackingField(WKing, Black)
  else
    if AttackingField(WKing, Black) = true then
      AssessPosition = MateValue - Depth
      exit function
    end if
    InCheck = AttackingField(BKing, White)
  end if
  Value = MaterialBalance(Depth)
  MatNum = MaterialTotal(Depth)
  if MatNum > MatB + MatB then
    if (Value < AAlpha - MaxPos) or (Value > ABeta + MaxPos) then
      AssessPosition = Value
      exit function
    end if
  end if
  InitAssessment
  PosValue = 0
  bBishlop = 0
  wBishlop = 0
  PawnCount = 0
  wRookon7 = 0
  bRookon2 = 0
  wDeveloped = Castling(White + 1)
  bDeveloped = Castling(Black + 1)
  if MoveControl(B1) > 0 then wDeveloped = wDeveloped + 1
  if MoveControl(C1) > 0 then wDeveloped = wDeveloped + 1
  if MoveControl(F1) > 0 then wDeveloped = wDeveloped + 1
  if MoveControl(G1) > 0 then wDeveloped = wDeveloped + 1
  if MoveControl(B8) > 0 then bDeveloped = bDeveloped + 1
  if MoveControl(C8) > 0 then bDeveloped = bDeveloped + 1
  if MoveControl(F8) > 0 then bDeveloped = bDeveloped + 1
  if MoveControl(G8) > 0 then bDeveloped = bDeveloped + 1
  for i = Column1 to Column8
    Feld = i * 10
    for j = RowA to HRow
      Feld = Feld + 1
      select case Board(Feld)
        case BK
          if MatNum < EndgameMaterial then
            PosValue = PosValue - CenterTable(Feld)
          else
            if Castling(Black + 1) = false then
              if (MoveControl(E8) > 0) or (MoveControl(H8) > 0) and (MoveControl(A8) > 0) then
                PosValue = PosValue + 35
              end if
            end if
            PosValue = PosValue + 4 * CenterTable(Feld)
            for k = -1 to 1
              if Board(Feld - 10 + k) = BP then PosValue = PosValue - 15
              if Board(Feld - 20 + k) = BP then PosValue = PosValue - 6
              if Pawns(j + k).White = 0 and Rooks(j + k).White > 0 then
                PosValue = PosValue + 12
              end if
            next k
          end if
        case BQ
          if bDeveloped < 4 then
            if Feld < A8 then PosValue = PosValue + 15
          else
            ColumnnDiff = abs(WKing \ 10 - Feld \ 10)
            RownDiff = abs(WKing mod 10 - Feld mod 10)
            PosValue = PosValue + 2 * (ColumnnDiff + RownDiff)
          end if
        case BN
          PosValue = PosValue - CenterTable(Feld) / 2
        case BB
          if(Feld = D6 or Feld = E6) and Board(Feld + 10) = BP then
            PosValue = PosValue + 20
          end if
          bBishlop = bBishlop + 1
        case BR
          if Feld <= H2 then bRookon2 = bRookon2 + 1
          if j >= CRow and j <= ERow then PosValue = PosValue - 4
          if Pawns(j).White = 0 then
            PosValue = PosValue - 8
            if Pawns(j).Black = 0 then PosValue = PosValue - 5
          end if
        case BP
          PosValue = PosValue - BPAssessment(Feld, i, j, bDeveloped)
          PawnCount = PawnCount + 1
        case Nil
        case WP
          PosValue = PosValue + WPAssessment(Feld, i, j, wDeveloped)
          PawnCount = PawnCount + 1
        case WR
          if Feld >= A7 then wRookon7 = wRookon7 + 1
          if j >= CRow and j <= ERow then PosValue = PosValue + 4
          if Pawns(j).Black = 0 then
            PosValue = PosValue + 8
            if Pawns(j).White = 0 then PosValue = PosValue + 5
          end if
        case WB
          if(Feld = D3 or Feld = E3) and Board(Feld - 10) = WP then
            PosValue = PosValue - 20
          end if
          wBishlop = wBishlop + 1
        case WN
          PosValue = PosValue + CenterTable(Feld) \ 2
        case WQ
          if wDeveloped < 4 then
            if Feld > H1 then PosValue = PosValue - 15
          else
            ColumnnDiff = abs(BKing \ 10 - Feld \ 10)
            RownDiff = abs(BKing mod 10 - Feld mod 10)
            PosValue = PosValue - 2 *(ColumnnDiff + RownDiff)
          end if
        case WK
          if MatNum < EndgameMaterial then
            PosValue = PosValue + CenterTable(Feld)
            if abs(Feld - BKing) = 20 or abs(Feld - BKing) = 2 then
              k = 10
              if MatNum = 0 then k = 30
              if SideToMove = White then
                PosValue = PosValue - k
              else
                PosValue = PosValue + k
              end if
            end if
          else
            if Castling(White + 1) = false then
              if MoveControl(E1) > 0 or(MoveControl(H1) > 0 and MoveControl(A1) > 0) then
                PosValue = PosValue - 35
              end if
            end if
            PosValue = PosValue - 4 * CenterTable(Feld)
            for k = -1 to 1
              if Board(Feld + 10 + k) = WP then PosValue = PosValue + 15
              if Board(Feld + 20 + k) = WP then PosValue = PosValue + 6
              if Pawns(j + k).Black = 0 and Rooks(j + k).Black > 0 then
                PosValue = PosValue - 12
              end if
            next k
          end if
      end select
    next j
  next i
  if PawnCount = 0 then
    Bed1 = MatNum <= MatB
    Bed2 = MatNum = 2 * MatN
    Bed3 = (MatNum <= 2 * MatB) and (abs(MaterialBalance(Depth)) < MatP)
    if Bed1 or Bed2 or Bed3 then
      AssessPosition = 0
      exit function
    end if
  end if
  if wBishlop >= 2 then PosValue = PosValue + 15
  if bBishlop >= 2 then PosValue = PosValue - 15
  if wRookon7 > 0 and BKing >= A7 then
    PosValue = PosValue + 10
    if wRookon7 > 1 then PosValue = PosValue + 25
  end if
  if bRookon2 > 0 and WKing <= H2 then
    PosValue = PosValue - 10
    if bRookon2 > 1 then PosValue = PosValue - 25
  end if
  if Side = Black then
    PosValue = -PosValue
  end if
  if Depth >= 1 then
    PosValue = PosValue -((Mobility(Depth) - Mobility(Depth - 1)) / 16)
  end if
  AssessPosition = Value + PosValue
end function

function AttackingField(AField as short, Side as short)as short
  dim as short i, Direction, LTo, Piece, slide
  if Side = White then
    if Board(AField - 9) = WP or Board(AField - 11) = WP then
      AttackingField = true
      exit function
    end if
  end if
  if Side = Black then
    if Board(AField + 9) = BP or Board(AField + 11) = BP then
      AttackingField = true
      exit function
    end if
  end if
  for i = 8 to 15
    LTo = AField + GOffset(i)
    if Board(LTo) = Nil or Board(LTo) = Edge then goto w1
    if Side = White then
      if Board(LTo) = WN then
        AttackingField = true
        exit function
      end if
    elseif Board(LTo) = BN then
      AttackingField = true
      exit function
    end if
    w1:
  next i
  for i = 0 to 7
    LTo = AField
    Direction = GOffset(i)
    Slide = 0
    Slideon1:
    Slide = Slide + 1
    LTo = LTo + Direction
    if Board(LTo) = Nil then
      goto Slideon1
    end if
    if Board(LTo) = Edge then goto w2
    Piece = Board(LTo)
    if Side = White then
      if Piece > 0 then
        if Piece = WK then
          if Slide <= 1 then
            AttackingField = true
            exit function
          end if
        else
          if FigOffset(Piece).Start <= i then
            if FigOffset(Piece).Ends >= i then
              AttackingField = true
              exit function
            end if
          end if
        end if
      end if
    else
      if Piece < 0 then
        if Piece = BK then
          if Slide <= 1 then
            AttackingField = true
            exit function
          end if
        else
          if FigOffset(-Piece).Start <= i then
            if FigOffset(-Piece).Ends >= i then
              AttackingField = true
              exit function
            end if
          end if
        end if
      end if
    end if
    w2:
  next i
  AttackingField = false
end function

function BPAssessment(AField as short, AColumn as short, ARow as short, ADeveloped as short) as short
  dim as short Value, Condition1, Condition2, j
  AColumn = (Column8 + Column1) - AColumn
  if MaterialTotal(Depth) > EndgameMaterial then
    Value = BPFieldValue(AField)
    if ADeveloped < 4 then
      if (ARow >= FRow or ARow <= BRow) and AColumn > Column3 then
        Value = Value - 15
      end if
    end if
  else
    Value = AColumn * 4
  end if
  if Pawns(ARow - 1).Black = 0 and Pawns(ARow + 1).Black = 0 then
    Value = Value - 12
    if Pawns(ARow).Black > 1 then Value = Value - 12
  end if
  if Pawns(ARow).Black > 1 then Value = Value - 15
  if PawnControlled(AField).Black > 0 or PawnControlled(AField - 10).Black > 0 then
    Value = Value + AColumn
  end if
  if Pawns(ARow).White = 0 then
    Condition1 = PawnControlled(AField).Black = 0
    Condition2 = PawnControlled(AField - 10).White > PawnControlled(AField - 10).Black
    if Condition1 and Condition2 then
      Value = Value - 10
      if Rooks(ARow).White > 0 then Value = Value - 8
    else
      for j = AField to A3 step -10
        if PawnControlled(j).White > 0 then
          BPAssessment = Value
          exit function
        end if
      next j
      if MaterialTotal(Depth) < EndgameMaterial then
        Value = Value + AColumn * 16
        if Rooks(ARow).Black > 0 then Value = Value + AColumn * 2
        if Rooks(ARow).White > 0 then Value = Value - AColumn * 2
        if MaterialTotal(Depth) = 0 then Value = Value + AColumn * 8
        if PawnControlled(AField).Black > 0 or PawnControlled(AField - 10).Black > 0 then
          Value = Value + AColumn * 4
        end if
        if Board(AField - 10) < 0 and PawnControlled(AField - 10).Black = 0 then
          Value = Value - AColumn * 4
        end if
      else
        Value = Value + AColumn * 8
        if PawnControlled(AField).Black > 0 or PawnControlled(AField - 10).Black > 0 then
          Value = Value + AColumn * 2
        end if
      end if
    end if
  end if
  BPAssessment = Value
end function

sub ComputerMove
  dim tmp as MoveType
  dim as short Value, Check, Distance, LAlpha, LBeta, MovesInLine, BestValue, i, j
  dim as integer Starttime, Endtime, Time_, Comtime
  dim as short LFrom, LTo
  dim LMove as string
  InitGameTree
  Value = AssessPosition(-1 * MateValue, MateValue, SideToMove)
  if Value = MateValue then
    Send("# Checkmate!")
    exit sub
  end if
  Check = InCheck
  NodeCount = 0
  Starttime = timer
  GenerateMoves(1)
  for Distance = 1 to MinDepth
    if Distance = 1 then
      LAlpha = -1 * MateValue
      LBeta = MateValue
    else
      LBeta = LAlpha + 100
      LAlpha = LAlpha - 100
    end if
    MaxExtension = Distance + 3
    MovesInLine = 0
    for i = 0 to StackLimit(1) - 1
      MovesInLine = MovesInLine + 1
      MVar(Depth, Depth).FromField = 0
      DisplayMove(i)
      if MovesInLine mod 9 = 8 then
        MovesInLine = 0
      end if
      PerformMove((i))
      Value = -1 * AlphaBeta(-1 * LBeta, -1 * LAlpha, Distance - 1)
      TakeBackMove((i))
      if i = 0 then
        if Value < LAlpha then
          LAlpha = -MateValue
          LBeta = Value
          MovesInLine = 0
          PerformMove(i)
          Value = -1 * AlphaBeta(-1 * LBeta, -1 * LAlpha, Distance - 1)
          TakeBackMove((i))
        elseif Value >= LBeta then
          LAlpha = Value
          LBeta = MateValue
          MovesInLine = 0
          PerformMove(i)
          Value = -1 * AlphaBeta(-1 * LBeta, -1 * LAlpha, Distance - 1)
          TakeBackMove(i)
        end if
        LAlpha = Value
        LBeta = LAlpha + 1
        DisplayMove(i)
        CopyMainVariant(i)
        MovesInLine = 0
      else
        if Value > LAlpha then
          BestValue = LAlpha
          LAlpha = Value
          LBeta = MateValue
          PerformMove(i)
          Value = -1 * AlphaBeta(-1 * LBeta, -1 * LAlpha, Distance - 1)
          TakeBackMove(i)
          if Value > BestValue then
            LAlpha = Value
            LBeta = LAlpha + 1
            CopyMainVariant(i)
            MovesInLine = 0
            tmp = MoveStack(i)
            for j = i to 1 step -1
              MoveStack(j) = MoveStack(j - 1)
            next j
            MoveStack(0) = tmp
          end if
        end if
      end if
    next i
  next Distance
  Endtime = timer
  if LAlpha > -1 * (MateValue - 1) then
    DisplayMove(0)
    Time_ = Endtime - Starttime
    Comtime = Comtime + Time_
    if Time_ = 0 then Time_ = 1
    PerformMove(0)
    
    LFrom = MoveStack(0).FromField
    LTo = MoveStack(0).ToField
    LMove = FieldNotation(LFrom) & FieldNotation(LTo)
    if MoveStack(0).PromotedPiece <> Nil then
      LMove = LMove & lcase(FigSymbol(MoveStack(0).PromotedPiece))
    end if
    
    Send("move " & LMove)
    
    if LAlpha >= MateValue - 10 then
      Send("# I checkmate in " & (MateValue - 2 - LAlpha) \ 2 & " moves.")
    else
      if LAlpha <= -MateValue + 10 then
        Send("# I'm checkmated in " & (LAlpha + MateValue - 1) \ 2 & " moves.")
      end if
    end if
  else
    if Check = true then
      Send("# Congratulations: MiniMAX is checkmated!")
    else
      Send("# Stalemate!")
    end if
  end if
end sub

sub CopyMainVariant(CurrMove as short)
  dim as short i
  MVar(Depth, Depth).FromField = MoveStack(CurrMove).FromField
  MVar(Depth, Depth).ToField = MoveStack(CurrMove).ToField
  i = 0
  do
    i = i + 1
    MVar(Depth, Depth + i) = MVar(Depth + 1, Depth + i)
  loop until MVar(Depth + 1, Depth + i).FromField = 0
end sub

function DisplayBoard() as string
  #ifdef __FB_LINUX__
  const EOL = chr(10)
  #else
  const EOL = chr(13, 10)
  #endif
  dim as short LColumn, LRow, LPiece, LSide
  dim result as string = ""
  for LColumn = Column8 to Column1 step -1
    result = result & "# "
    for LRow = RowA to HRow
      LPiece = Board(LColumn * 10 + LRow)
      LSide = sgn(LPiece)
      LPiece = abs(LPiece)
      if LSide = Black then
        result = result & lcase(FigSymbol(LPiece)) & " "
      else
        result = result & FigSymbol(LPiece) & " "
      end if
    next LRow
    result = result & EOL
  next LColumn
  if SideToMove = White then
    result = result & "# White"
  else
    result = result & "# Black"
  end if
  result = result & " to move" & EOL
  result = result & "# Material balance: " & MaterialBalance(Depth) & EOL
  result = result & "# En Passant field: " & FieldNotation(EpField(Depth)) & EOL
  result = result & "# Castling state black: "
  if MoveControl(E8) + MoveControl(H8) = 0 then result = result & "0-0 "
  if MoveControl(E8) + MoveControl(A8) = 0 then result = result & "0-0-0"
  result = result & EOL
  result = result & "# Castling state white: "
  if MoveControl(E1) + MoveControl(H1) = 0 then result = result & "0-0 "
  if MoveControl(E1) + MoveControl(A1) = 0 then result = result & "0-0-0"
  return result
end function

sub DisplayMove(CurrMove as short)
  dim as short LFrom, LTo
  LFrom = MoveStack(CurrMove).FromField
  LTo = MoveStack(CurrMove).ToField
  if MoveStack(CurrMove).CapturedPiece = Nil then
  end if
  if MoveStack(CurrMove).PromotedPiece <> Nil then
  end if
end sub

function FieldNotation(AField as short) as string
  dim as integer LCol, LRow
  if (AField < A1) or (AField > H8) or (Board(AField) = Edge) then
    FieldNotation = "--"
  else
    LCol = AField mod 10
    LRow = AField \ 10 - 1
    FieldNotation = chr(asc("a") + LCol - 1, asc("1") + LRow - 1)
  end if
end function

function FieldNumber(AField as string) as short
  dim as string LRow = left(AField, 1)
  dim as string LCol = mid(AField, 2, 1)
  if (LRow < "a") or (LRow > "h") or (LCol < "1") or (LCol > "8") then
    FieldNumber = Illegal
    exit function
  end if
  FieldNumber = 10 * (asc(LCol) - asc("1") + 2) + asc(LRow) - asc("a") + 1
end function

sub GenerateMoves(AllMoves as short)
  dim as short LFrom, LPiece, Longpaths, i, Direction, LTo, CaptureMove, ep, OK
  Index = StackLimit(Depth)
  Mobility(Depth) = 0
  for LFrom = A1 to H8
    LPiece = Board(LFrom)
    if (LPiece = Nil) or (LPiece = Edge) then goto NextField
    if (SideToMove = White) and (LPiece < 0) then goto NextField
    if (SideToMove = Black) and (LPiece > 0) then goto NextField
    LPiece = abs(LPiece)
    if LPiece = WP then
      if SideToMove = White then
        if Board(LFrom + 10) = Nil then
          if LFrom >= A7 then
            SavePromotion(LFrom, LFrom + 10)
          elseif AllMoves > 0 then
            SaveMove(LFrom, LFrom + 10)
            if LFrom <= H2 and Board(LFrom + 20) = Nil then
              SaveMove(LFrom, LFrom + 20)
              MoveStack(Index - 1).EpField = LFrom + 10
            end if
          end if
        end if
        if Board(LFrom + 11) < 0 then
          if LFrom >= A7 then
            SavePromotion(LFrom, LFrom + 11)
          else
            SaveCaptureMove(LFrom, LFrom + 11)
          end if
        end if
        if Board(LFrom + 9) < 0 then
          if LFrom >= A7 then
            SavePromotion(LFrom, LFrom + 9)
          else
            SaveCaptureMove(LFrom, LFrom + 9)
          end if
        end if
      elseif SideToMove = Black then
        if Board(LFrom - 10) = Nil then
          if LFrom <= H2 then
            SavePromotion(LFrom, LFrom - 10)
          elseif AllMoves > 0 then
            SaveMove(LFrom, LFrom - 10)
            if LFrom >= A7 and Board(LFrom - 20) = Nil then
              SaveMove(LFrom, LFrom - 20)
              MoveStack(Index - 1).EpField = LFrom - 10
            end if
          end if
        end if
        if Board(LFrom - 11) > 0 and Board(LFrom - 11) <> Edge then
          if LFrom <= H2 then
            SavePromotion(LFrom, LFrom - 11)
          else
            SaveCaptureMove(LFrom, LFrom - 11)
          end if
        end if
        if Board(LFrom - 9) > 0 and Board(LFrom - 9) <> Edge then
          if LFrom <= H2 then
            SavePromotion(LFrom, LFrom - 9)
          else
            SaveCaptureMove(LFrom, LFrom - 9)
          end if
        end if
      end if
      goto NextField
    end if
    Longpaths = FigOffset(LPiece).Longpaths
    for i = FigOffset(LPiece).Start to FigOffset(LPiece).Ends
      Direction = GOffset(i)
      LTo = LFrom
      Slideon2:
      LTo = LTo + Direction
      if Board(LTo) = Nil then
        if AllMoves > 0 then
          SaveMove(LFrom, LTo)
        end if
        if Longpaths then
          goto Slideon2
        else
          goto NextDirection
        end if
      end if
      if Board(LTo) = Edge then
        goto NextDirection
      end if
      CaptureMove = (SideToMove = White) and (Board(LTo) < 0)
      CaptureMove = CaptureMove or (SideToMove = Black) and (Board(LTo) > 0)
      if CaptureMove then SaveCaptureMove(LFrom, LTo)
      NextDirection:
    next i
    NextField:
  next LFrom
  if EpField(Depth) <> Illegal then
    ep = EpField(Depth)
    if SideToMove = White then
      if Board(ep - 9) = WP then
        SaveEnPassantMove(ep - 9, ep, ep - 10)
      end if
      if Board(ep - 11) = WP then
        SaveEnPassantMove(ep - 11, ep, ep - 10)
      end if
    else
      if Board(ep + 9) = BP then
        SaveEnPassantMove(ep + 9, ep, ep + 10)
      end if
      if Board(ep + 11) = BP then
        SaveEnPassantMove(ep + 11, ep, ep + 10)
      end if
    end if
  end if
  if SideToMove = White then
    if WKing = E1 and MoveControl(E1) = 0 then
      OK = Board(H1) = WR and MoveControl(H1) = 0
      OK = OK and Board(F1) = Nil and Board(G1) = Nil
      OK = OK and AttackingField(E1, Black) = false
      OK = OK and AttackingField(F1, Black) = false
      OK = OK and AttackingField(G1, Black) = false
      if OK then
        SaveMove(E1, G1)
        MoveStack(Index - 1).CastlingNr = ShortCastlingMove
      end if
      OK = Board(A1) = WR and MoveControl(A1) = 0
      OK = OK and Board(D1) = Nil
      OK = OK and Board(C1) = Nil
      OK = OK and Board(B1) = Nil
      OK = OK and AttackingField(E1, Black) = false
      OK = OK and AttackingField(D1, Black) = false
      OK = OK and AttackingField(C1, Black) = false
      if OK then
        SaveMove(E1, C1)
        MoveStack(Index - 1).CastlingNr = LongCastlingMove
      end if
    end if
  else
    if BKing = E8 and MoveControl(E8) = 0 then
      OK = Board(H8) = BR and MoveControl(H8) = 0
      OK = OK and Board(F8) = Nil and Board(G8) = Nil
      OK = OK and AttackingField(E8, White) = false
      OK = OK and AttackingField(F8, White) = false
      OK = OK and AttackingField(G8, White) = false
      if OK then
        SaveMove(E8, G8)
        MoveStack(Index - 1).CastlingNr = ShortCastlingMove
      end if
      OK = Board(A8) = BR and MoveControl(A8) = 0
      OK = OK and Board(D8) = Nil
      OK = OK and Board(C8) = Nil
      OK = OK and Board(B8) = Nil
      OK = OK and AttackingField(E8, White) = false
      OK = OK and AttackingField(D8, White) = false
      OK = OK and AttackingField(C8, White) = false
      if OK then
        SaveMove(E8, C8)
        MoveStack(Index - 1).CastlingNr = LongCastlingMove
      end if
    end if
  end if
  StackLimit(Depth + 1) = Index
end sub

sub InitAssessment
  dim i as short
  for i = A1 to H8
    PawnControlled(i).White = 0
    PawnControlled(i).Black = 0
  next i
  for i = RowA - 1 to HRow + 1
    Pawns(i).White = 0
    Pawns(i).Black = 0
    Rooks(i).White = 0
    Rooks(i).Black = 0
  next i
  for i = A1 to H8
    if (Board(i) = Nil) or (Board(i) = Edge) then goto NextField
    select case Board(i)
      case WP
        PawnControlled(i + 9).White = PawnControlled(i + 9).White + 1
        PawnControlled(i + 11).White = PawnControlled(i + 11).White + 1
        Pawns(i mod 10).White = Pawns(i mod 10).White + 1
      case BP
        PawnControlled(i - 9).Black = PawnControlled(i - 9).Black + 1
        PawnControlled(i - 11).Black = PawnControlled(i - 11).Black + 1
        Pawns(i mod 10).Black = Pawns(i mod 10).Black + 1
      case BR
        Rooks(i mod 10).Black = Rooks(i mod 10).Black + 1
      case WR
        Rooks(i mod 10).White = Rooks(i mod 10).White + 1
      case else
    end select
    NextField:
  next i
end sub

sub InitGameTree
  if Depth = 0 then exit sub
  EpField(0) = EpField(1)
  MaterialBalance(0) = MaterialBalance(1)
  MaterialTotal(0) = MaterialTotal(1)
  Depth = 0
end sub

function InputMove(Move as string) as short
  dim as short LFrom, LTo, i, tmp
  if len(Move) < 4 then
    InputMove = false
    exit function
  end if
  LFrom = FieldNumber(Move)
  LTo = FieldNumber(mid(Move, 3, 2))
  GenerateMoves(1)
  for i = StackLimit(Depth) to StackLimit(Depth + 1) - 1
    if MoveStack(i).FromField = LFrom and MoveStack(i).ToField = LTo then
      if MoveStack(i).PromotedPiece <> Nil then
        if mid(Move, 5, 1) = "n" then
          i = i + 1
        elseif mid(Move, 5, 1) = "b" then
          i = i + 2
        elseif mid(Move, 5, 1) = "r" then
          i = i + 3
        end if
      end if
      InitGameTree
      DisplayMove(i)
      tmp = LastMove
      PerformMove(i)
      if SideToMove = Black then
        if AttackingField(WKing, Black) then
          LogLn("** White king on " & FieldNotation(WKing) & " is being checked.")
          TakeBackMove((i))
          LastMove = tmp
          InputMove = false
          exit function
        end if
      elseif AttackingField(BKing, White) then
        LogLn("** Black king on " & FieldNotation(BKing) & " is being checked.")
        TakeBackMove((i))
        LastMove = tmp
        InputMove = false
        exit function
      end if
      InputMove = true
      exit function
    end if
  next i
  InputMove = false
end function

function NextBestMove() as short
  dim as short BestMove, BestValue, i
  BestMove = -1
  BestValue = -1 * MateValue
  for i = StackLimit(Depth) to StackLimit(Depth + 1) - 1
    if MoveStack(i).Value > BestValue then
      BestMove = i
      BestValue = MoveStack(i).Value
    end if
  next i
  if BestMove >= 0 then MoveStack(BestMove).Value = -1 * MateValue
  NextBestMove = BestMove
end function

sub PerformMove(CurrMove as short)
  dim as short LFrom, LTo, LEnPassant, MatChange
  MoveCount += 1
  LFrom = MoveStack(CurrMove).FromField
  LTo = MoveStack(CurrMove).ToField
  LEnPassant = MoveStack(CurrMove).EpField
  LastMove = CurrMove
  Depth += 1
  ToField(Depth) = LTo
  EpField(Depth) = Illegal
  MaterialBalance(Depth) = -1 * MaterialBalance(Depth - 1)
  MaterialTotal(Depth) = MaterialTotal(Depth - 1)
  MoveControl(LFrom) = MoveControl(LFrom) + 1
  MoveControl(LTo) = MoveControl(LTo) + 1
  if LEnPassant <> Illegal then
    if Board(LEnPassant) = Nil then
      EpField(Depth) = LEnPassant
    else
      Board(LEnPassant) = Nil
      MaterialBalance(Depth) = MaterialBalance(Depth) - MatP
    end if
  else
    if MoveStack(CurrMove).CapturedPiece <> Nil then
      MatChange = PieceMaterial(MoveStack(CurrMove).CapturedPiece)
      MaterialBalance(Depth) = MaterialBalance(Depth) - MatChange
      if MatChange <> MatP then
        MaterialTotal(Depth) = MaterialTotal(Depth) - MatChange
      end if
    end if
  end if
  Board(LTo) = Board(LFrom)
  Board(LFrom) = Nil
  if MoveStack(CurrMove).PromotedPiece <> Nil then
    Board(LTo) = SideToMove * MoveStack(CurrMove).PromotedPiece
    MatChange = PieceMaterial(MoveStack(CurrMove).PromotedPiece) - MatP
    MaterialBalance(Depth) = MaterialBalance(Depth) - MatChange
    MaterialTotal(Depth) = MaterialTotal(Depth) + MatChange + MatP
  else
    if MoveStack(CurrMove).CastlingNr = ShortCastlingMove then
      Board(LTo + 1) = Nil
      Board(LTo - 1) = SideToMove * WR
      Castling(SideToMove + 1) = true
    elseif MoveStack(CurrMove).CastlingNr = LongCastlingMove then
      Board(LTo - 2) = Nil
      Board(LTo + 1) = SideToMove * WR
      Castling(SideToMove + 1) = true
    end if
  end if
  if Board(LTo) = WK then
    WKing = LTo
  elseif Board(LTo) = BK then
    BKing = LTo
  end if
  SideToMove = -1 * SideToMove
end sub

sub SaveCaptureMove(AFrom as short, ATo as short)
  dim as short FigValue, Killer1, Killer2, MVarMove, i
  if Board(ATo) = WK or Board(ATo) = BK then exit sub
  FigValue = PieceMaterial(abs(Board(ATo)))
  MoveStack(Index).FromField = AFrom
  MoveStack(Index).ToField = ATo
  MoveStack(Index).CapturedPiece = abs(Board(ATo))
  MoveStack(Index).Value = FigValue - PieceMaterial(abs(Board(AFrom))) \ 8
  if Depth > 0 then
    if ATo = ToField(Depth - 1) then
      MoveStack(Index).Value += 300
    end if
  end if
  Killer1 = KillerTab(Depth).Killer1.FromField = AFrom
  Killer1 = Killer1 and KillerTab(Depth).Killer1.ToField = ATo
  Killer2 = KillerTab(Depth).Killer2.FromField = AFrom
  Killer2 = Killer2 and KillerTab(Depth).Killer2.ToField = ATo
  MVarMove = (MVar(0, Depth).FromField = AFrom) and (MVar(0, Depth).ToField = ATo)
  if MVarMove then
    MoveStack(Index).Value += MainVariantBonus
  elseif Killer1 then
    MoveStack(Index).Value += Killer1Bonus
  elseif Killer2 then
    MoveStack(Index).Value += Killer2Bonus
  end if
  MoveStack(Index).PromotedPiece = Nil
  MoveStack(Index).CastlingNr = NoCastlingMove
  MoveStack(Index).EpField = Illegal
  if Index < CMoveStackDim then
    Index += 1
  else
    print("ERROR: Move stack overflow")
    system
  end if
end sub

sub SaveEnPassantMove(AFrom as short, ATo as short, AEP as short)
  if Board(ATo) = WK or Board(ATo) = BK then exit sub
  MoveStack(Index).FromField = AFrom
  MoveStack(Index).ToField = ATo
  MoveStack(Index).CapturedPiece = WP
  MoveStack(Index).PromotedPiece = Nil
  MoveStack(Index).CastlingNr = NoCastlingMove
  MoveStack(Index).EpField = AEP
  MoveStack(Index).Value = MatP
  if Index < CMoveStackDim then
    Index += 1
  else
    print("ERROR: Move stack overflow")
    system
  end if
end sub

sub SaveMove(AFrom as short, ATo as short)
  dim as short Killer1, Killer2, MVarMove
  if SideToMove = White then
    if Board(AFrom) = WB or Board(AFrom) = WR then
      Mobility(Depth) = Mobility(Depth) + CenterTable(ATo)
    end if
  else
    if Board(AFrom) = BB or Board(AFrom) = BR then
      Mobility(Depth) = Mobility(Depth) + CenterTable(ATo)
    end if
  end if
  Killer1 = KillerTab(Depth).Killer1.FromField = AFrom
  Killer1 = Killer1 and KillerTab(Depth).Killer1.ToField = ATo
  Killer2 = KillerTab(Depth).Killer2.FromField = AFrom
  Killer2 = Killer2 and KillerTab(Depth).Killer2.ToField = ATo
  MVarMove = MVar(0, Depth).FromField = AFrom and MVar(0, Depth).ToField = ATo
  if MVarMove then
    MoveStack(Index).Value = MainVariantBonus
  elseif Killer1 then
    MoveStack(Index).Value = Killer1Bonus
  elseif Killer2 then
    MoveStack(Index).Value = Killer2Bonus
  else
    MoveStack(Index).Value = Nil
  end if
  MoveStack(Index).FromField = AFrom
  MoveStack(Index).ToField = ATo
  MoveStack(Index).CapturedPiece = Nil
  MoveStack(Index).PromotedPiece = Nil
  MoveStack(Index).CastlingNr = NoCastlingMove
  MoveStack(Index).EpField = Illegal
  if Index < CMoveStackDim then
    Index += 1
  else
    print " ERROR: Move stack overflowed"
    system
  end if
end sub

sub SavePromotion(AFrom as short, ATo as short)
  dim i as short
  if Board(ATo) = Nil then
    for i = WQ to WR step -1
      SaveMove(AFrom, ATo)
      MoveStack(Index - 1).PromotedPiece = i
    next i
  else
    for i = WQ to WR step -1
      SaveCaptureMove(AFrom, ATo)
      MoveStack(Index - 1).PromotedPiece = i
    next i
  end if
end sub

sub TakeBackMove(CurrMove as short)
  dim as short LFrom, LTo, LEnPassant
  MoveCount = MoveCount - 1
  LFrom = MoveStack(CurrMove).FromField
  LTo = MoveStack(CurrMove).ToField
  LEnPassant = MoveStack(CurrMove).EpField
  SideToMove = -1 * SideToMove
  Depth -= 1
  Board(LFrom) = Board(LTo)
  Board(LTo) = Nil
  if (LEnPassant <> Illegal) and (MoveStack(CurrMove).CapturedPiece = WP) then
    Board(LEnPassant) = -1 * SideToMove
  elseif MoveStack(CurrMove).CapturedPiece <> Nil then
    Board(LTo) = -1 * SideToMove * MoveStack(CurrMove).CapturedPiece
  end if
  MoveControl(LFrom) -= 1
  MoveControl(LTo) -= 1
  if MoveStack(CurrMove).CastlingNr = ShortCastlingMove then
    Board(LTo + 1) = SideToMove * WR
    Board(LTo - 1) = Nil
    Castling(SideToMove + 1) = false
  elseif MoveStack(CurrMove).CastlingNr = LongCastlingMove then
    Board(LTo - 2) = SideToMove * WR
    Board(LTo + 1) = Nil
    Castling(SideToMove + 1) = false
  end if
  if MoveStack(CurrMove).PromotedPiece <> Nil then
    Board(LFrom) = SideToMove
  end if
  if Board(LFrom) = WK then
    WKing = LFrom
  elseif Board(LFrom) = BK then
    BKing = LFrom
  end if
end sub

function WPAssessment(AField as short, AColumn as short, ARow as short, ADeveloped as short)as short
  dim as short Value, Condition1, Condition2, j
  if MaterialTotal(Depth) > EndgameMaterial then
    Value = WPFieldValue(AField)
    if ADeveloped < 4 then
      if ((ARow >= FRow) or (ARow <= BRow)) and (AColumn > Column3) then
        Value = Value - 15
      end if
    end if
  else
    Value = AColumn * 4
  end if
  if (Pawns(ARow - 1).White = 0) and (Pawns(ARow + 1).White = 0) then
    Value -= 12
    if Pawns(ARow).White > 1 then Value -= 12
  end if
  if Pawns(ARow).White > 1 then Value -= 15
  if (PawnControlled(AField).White > 0) or (PawnControlled(AField + 10).White > 0) then
    Value += AColumn
  end if
  if Pawns(ARow).Black = 0 then
    Condition1 = PawnControlled(AField).White = 0
    Condition2 = PawnControlled(AField + 10).Black > PawnControlled(AField + 10).White
    if Condition1 and Condition2 then
      Value = Value - 10
      if Rooks(ARow).Black > 0 then Value = Value - 8
    else
      for j = AField to H6 step 10
        if PawnControlled(j).Black > 0 then
          WPAssessment = Value
          exit function
        end if
      next j
      if MaterialTotal(Depth) < EndgameMaterial then
        Value = Value + AColumn * 16
        if Rooks(ARow).White > 0 then Value += AColumn * 2
        if Rooks(ARow).Black > 0 then Value -= AColumn * 2
        if MaterialTotal(Depth) = 0 then Value += AColumn * 8
        if (PawnControlled(AField).White > 0) or (PawnControlled(AField + 10).White > 0) then
          Value += AColumn * 4
        end if
        if (Board(AField + 10) < 0) and (PawnControlled(AField + 10).White = 0) then
          Value -= AColumn * 4
        end if
      else
        Value += AColumn * 8
        if (PawnControlled(AField).White > 0) or (PawnControlled(AField + 10).White > 0) then
          Value += AColumn * 2
        end if
      end if
    end if
  end if
  WPAssessment = Value
end function
